#ifndef _ADC_H__
#define _ADC_H__

#include <stdint.h>
#include <avr/io.h>


/*
Use internal ADC

Missing features: input-gain, differential channels, digital input disable

Target:      ATmega168
Last Change: 2015-04-18 (Frank Baettermann)
*/


// According to manual the adc-clock should be below 200 kHz
#define ADC_CLOCK_MAX 200000

// Select prescaler for maximum adc-frequency
#if (F_CPU < ADC_CLOCK_MAX)
    #define ADC_PRESCALER       1
    #define ADC_PRESCALER_REG   0
#elif ((F_CPU / 2) <= ADC_CLOCK_MAX)
    #define ADC_PRESCALER       2
    #define ADC_PRESCALER_REG   (1 << ADPS0);
#elif ((F_CPU / 4) <= ADC_CLOCK_MAX)
    #define ADC_PRESCALER       4
    #define ADC_PRESCALER_REG   (1 << ADPS1)
#elif ((F_CPU / 8) <= ADC_CLOCK_MAX)
    #define ADC_PRESCALER       8
    #define ADC_PRESCALER_REG   (1 << ADPS1 | 1 << ADPS0)
#elif ((F_CPU / 16) <= ADC_CLOCK_MAX)
    #define ADC_PRESCALER       16
    #define ADC_PRESCALER_REG   (1 << ADPS2)
#elif ((F_CPU / 32) <= ADC_CLOCK_MAX)
    #define ADC_PRESCALER       32
    #define ADC_PRESCALER_REG   (1 << ADPS2 | 1 << ADPS0)
#elif ((F_CPU / 64) <= ADC_CLOCK_MAX)
    #define ADC_PRESCALER       64
    #define ADC_PRESCALER_REG   (1 << ADPS2 | 1 << ADPS1)
#elif ((F_CPU / 128) <= ADC_CLOCK_MAX)
    #define ADC_PRESCALER       128
    #define ADC_PRESCALER_REG   (1 << ADPS2 | 1 << ADPS1 | 1 << ADPS0)
#else
    #error "Reduce F_CPU to keep the ADC-clock below 200kHz."
#endif

#define ADC_CLOCK               (F_CPU / ADC_PRESCALER)
#define ADC_REFVOLTAGE_AREF     2500UL  // [mV] External reference voltage
#define ADC_REFVOLTAGE_INT      1100UL  // [mV]
#define ADC_REFVOLTAGE_AVCC     5000UL  // [mV]
#define ADC_MAX_VAL             1023UL  // 10 bit

// ADC supersampling (~12 bit)
#define ADC_SS_COUNT            64 // 16 should do, but we're not in a hurry
#define ADC_SS_DIV              (ADC_SS_COUNT / 4)
#define ADC_SS_MAX              (ADC_MAX_VAL * 4)



void adc_init(void);


static inline void adc_set_channel(uint8_t channel)
{
    // 0-7      single ended channels
    //   8      internal temperature reference
    //  14      internal voltage reference
    //  15      GND
    
    // As the MUX3:0-bits are right aligned in ADMUX-register we only mask them
    ADMUX = (ADMUX & 0xf0) | (channel & 0x0f);
};


static inline void adc_enable(void)
{
    ADCSRA |= (1 << ADEN);
};


static inline void adc_disable(void)
{
    ADCSRA &= ~(1 << ADEN);
};


static inline void adc_start_conversion(void)
{
    ADCSRA |= (1 << ADSC);
}


// Do not call from outside
static inline uint16_t _adc_conversion(void)
{
    adc_start_conversion();
    while (ADCSRA & (1 << ADSC)); // loop until the conversion is complete
    return ADCW;
};


#define ADC_CHANNEL_PANEL_VOLTAGE     0
#define ADC_CHANNEL_PANEL_CURRENT     1
#define ADC_CHANNEL_LOAD_CURRENT      2
#define ADC_CHANNEL_BATT_VOLTAGE      3

extern volatile uint16_t adc_voltage;
extern volatile uint8_t adc_voltage_new;
extern volatile uint16_t adc_current;
extern volatile uint8_t adc_current_new;
extern volatile uint16_t adc_battery_voltage;
extern volatile uint8_t adc_battery_voltage_new;
extern volatile uint16_t adc_load_current;
extern volatile uint8_t adc_load_current_new;


//uint16_t adc_conversion(uint8_t channel);
//uint16_t adc_12BitConversion(uint8_t channel);

#endif //_ADC_H__
