#include <stdint.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "hd44780.h"

#ifdef HD44780_MODE_PCF8574
#include "pcf8574.h"

// Current port expander setting
uint8_t pcf8574 = 0;
// Backlight setting
uint8_t pcf8574_bl = HD44780_PCF8574_BL;

#endif // HD44780_MODE_PCF8574

// Initialize LCD
// Returns 0 if successful, -1 otherwise.
int8_t hd44780_init(void)
{
#ifdef HD44780_MODE_8BIT
    // Set port pins to output
    HD44780_DDR |= (1 << HD44780_RS) | (1 << HD44780_RW) | (1 << HD44780_EN);
    HD44780_DDDR = 0xff;
#endif // HD44780_MODE_8BIT
#ifdef HD44780_MODE_PCF8574
    // Let port expander initialize
    _delay_ms(20);
    // All pins low
    if (hd44780_pcf8574_write(0))
        goto error;
#endif // HD44780_MODE_PCF8574

    // LCD initializes itself...
    _delay_ms(20);

#ifdef HD44780_MODE_8BIT
     // Init sequence: all low except DB4, DB5
    HD44780_PORT &= ~((1 << HD44780_RS) | (1 << HD44780_RW) | (1 << HD44780_EN));
    HD44780_DPORT = 0x30;
#endif // HD44780_MODE_8BIT
#ifdef HD44780_MODE_PCF8574
    pcf8574_write(HD44780_PCF8574_I2CADDR, HD44780_PCF8574_DB4 | HD44780_PCF8574_DB5);
#endif // HD44780_MODE_PCF8574

     // Init sequence (datasheet...)
    hd44780_enablePulse();
    _delay_ms(5);
    hd44780_enablePulse();
    _delay_ms(1);
    hd44780_enablePulse();
    _delay_ms(1);

    // Function set
#ifdef HD44780_MODE_8BIT
    hd44780_write_cmd((1 << 5) | HD44780_CMD_FS_8BIT | HD44780_CMD_FS_2LINE | HD44780_CMD_FS_5X7FONT);
#endif // HD44780_MODE_8BIT
#ifdef HD44780_MODE_PCF8574
    hd44780_write_cmd((1 << 5) | HD44780_CMD_FS_4BIT | HD44780_CMD_FS_2LINE | HD44780_CMD_FS_5X7FONT);
#endif // HD44780_MODE_PCF8574
    // Cursor/display shift
    hd44780_write_cmd((1 << 4) | HD44780_CMD_CS_CURSORRIGHT);
    // Display control
    hd44780_write_cmd((1 << 3) | HD44780_CMD_DC_ENABLE);
    return 0;

    hd44780_clear();
error:
    return -1;
}


void hd44780_pcf8574_backlight_on(void)
{
    pcf8574_bl |= HD44780_PCF8574_BL;
    hd44780_pcf8574_write(pcf8574);
}


void hd44780_pcf8574_backlight_off(void)
{
    pcf8574_bl &= ~HD44780_PCF8574_BL;
    hd44780_pcf8574_write(pcf8574);
}


int8_t hd44780_pcf8574_write(uint8_t d)
{
    pcf8574 = d;
    return pcf8574_write(HD44780_PCF8574_I2CADDR, d | pcf8574_bl);
}


void hd44780_enablePulse(void)
{
#ifdef HD44780_MODE_8BIT
    HD44780_PORT |= (1 << HD44780_EN);
    _delay_us(1);
    HD44780_PORT &= ~(1 << HD44780_EN);
#endif // HD44780_MODE_8BIT
#ifdef HD44780_MODE_PCF8574
    hd44780_pcf8574_write(pcf8574 | HD44780_PCF8574_EN);
    // ... writing to bus introduces interval...
    hd44780_pcf8574_write(pcf8574 & ~HD44780_PCF8574_EN);
#endif // HD44780_MODE_PCF8574
}

void hd44780_write_cmd(char c)
{
    // RW low = write, RS low = command
#ifdef HD44780_MODE_8BIT
    HD44780_PORT &= ~((1 << HD44780_RW) | (1 << HD44780_RS));
    HD44780_DPORT = c;
    hd44780_enablePulse();
    hd44780_busyLoop();
#endif // HD44780_MODE_8BIT
#ifdef HD44780_MODE_PCF8574
    // Datasheet states up to 1.6ms execution time: use busy loop
    // Upper nibble
    hd44780_pcf8574_write((c & 0x80 ? HD44780_PCF8574_DB7 : 0) | (c & 0x40 ? HD44780_PCF8574_DB6 : 0) | (c & 0x20 ? HD44780_PCF8574_DB5 : 0) | (c & 0x10 ? HD44780_PCF8574_DB4 : 0));
    hd44780_enablePulse();
    hd44780_busyLoop();
    // Lower nibble
    hd44780_pcf8574_write((c & 0x08 ? HD44780_PCF8574_DB7 : 0) | (c & 0x04 ? HD44780_PCF8574_DB6 : 0) | (c & 0x02 ? HD44780_PCF8574_DB5 : 0) | (c & 0x01 ? HD44780_PCF8574_DB4 : 0));
    hd44780_enablePulse();
    hd44780_busyLoop();
#endif // HD44780_MODE_PCF8574
}


void hd44780_clear(void)
{
    hd44780_write_cmd(HD44780_CMD_CLEAR_DISPLAY);
    hd44780_write_cmd(HD44780_CMD_CURSOR_HOME);
}


// Write character
void hd44780_putc(char c)
{
    // RW low = write, RS high = data
#ifdef HD44780_MODE_8BIT
    HD44780_PORT = (HD44780_PORT & ~(1 << HD44780_RW)) | (1 << HD44780_RS);
    HD44780_DPORT = c;
    hd44780_enablePulse();
    hd44780_busyLoop();
#endif // HD44780_MODE_8BIT
#ifdef HD44780_MODE_PCF8574
    // Datasheet states 47µs execution time for writing to RAM: just use delay
    // Upper nibble
    hd44780_pcf8574_write(HD44780_PCF8574_RS | (c & 0x80 ? HD44780_PCF8574_DB7 : 0) | (c & 0x40 ? HD44780_PCF8574_DB6 : 0) | (c & 0x20 ? HD44780_PCF8574_DB5 : 0) | (c & 0x10 ? HD44780_PCF8574_DB4 : 0));
    hd44780_enablePulse();
    _delay_us(45);
    // Lower nibble
    hd44780_pcf8574_write(HD44780_PCF8574_RS | (c & 0x08 ? HD44780_PCF8574_DB7 : 0) | (c & 0x04 ? HD44780_PCF8574_DB6 : 0) | (c & 0x02 ? HD44780_PCF8574_DB5 : 0) | (c & 0x01 ? HD44780_PCF8574_DB4 : 0));
    hd44780_enablePulse();
    _delay_us(45);
#endif // HD44780_MODE_PCF8574
}


//// Convenience function: Write zero terminated string
//void hd44780_puts(char *s)
//{
//    while (*s)
//    {
//        hd44780_putc(*s++);
//    }
//}
//
//
//// Write zero terminated string from pgmspace
//void hd44780_puts_P(const char *s)
//{
//    for (;;)
//    {
//        char c = pgm_read_byte(s++);
//        if (!c)
//            break;
//        hd44780_putc(c);
//    }
//}


// Wait for the busy-flag (DB7) to become low
void hd44780_busyLoop(void)
{
#ifdef HD44780_MODE_8BIT
    // data-pins low and input
    HD44780_DPORT = 0x00;
    HD44780_DDDR = 0x00;
    // Set read-command
    HD44780_PORT &= ~(1 << HD44780_RS);
    HD44780_PORT |= (1 << HD44780_RW);
    do
    {
        hd44780_enablePulse(); // Read busy flag
    }
    while(HD44780_DPIN & 128);
    // Sata-pins as output
    HD44780_DDDR = 0xff;
#endif // HD44780_MODE_8BIT
#ifdef HD44780_MODE_PCF8574
    // Access via Bus is slow anyway. We just wait for
    // the LCD to complete the requested action.
    _delay_ms(2);
#endif // HD44780_MODE_PCF8574
}


void hd44780_set_cursor(uint8_t x, uint8_t y)
{
    uint8_t tmp = 0x80;
    switch (y)
    {
        case 0: tmp += x; break;        // 1st row
        case 1: tmp += 0x40 + x; break; // 2nd row
        case 2: tmp += 0x14 + x; break; // 3rd row
        case 3: tmp += 0x54 + x; break; // 4th row
    }
    hd44780_write_cmd(tmp);
}


// Define user configurable characters (5 by 8 pixels)
void hd44780_define_char(uint8_t no, uint8_t* data)
{
    // Set LCD address to CG RAM
    hd44780_write_cmd(0x40 + no*HD44780_USERCHAR_LENGTH);
    // Enter character bytes into CG RAM
    for (uint8_t i = 0; i < HD44780_USERCHAR_LENGTH; i++)
        hd44780_putc(data[i]);
}

