#include <stdint.h>
#include <avr/io.h>
#include <util/twi.h> 
#include "i2c.h"


/*
 * Control TWI-interface. Use external pullups.
 *
 * Last change: 2015-04-16 (Frank Baettermann)
 * Target: ATmega88
 *
 * Based on work by
 *      Peter Fleury (http://jump.to/fleury)
 *      Hannes Mayer (http://www.captain.at)
 */


void i2c_init(void)
{
    //                        F_CPU
    // SCL frequency = ---------------------------
    //                  16 + 2 * TWBR * prescaler
    //
    // - freq < 400kHz
    // - F_CPU should be 16 times higher than freq.

    TWSR |= I2C_PRESCALER;
    TWBR = I2C_TWBR;
}

// generic write function, used for both addressing and data (do NOT call it directly!)
// 
uint8_t i2c_write(uint8_t byte)
{
    TWDR = byte; // write byte to register
    TWCR = (1 << TWINT) | (1 << TWEN); // initiate transmission
    loop_until_bit_is_set(TWCR, TWINT); // wait
    return TW_STATUS;
}

// ===============

uint8_t i2c_start(void)
{
    /*
    - TWINT: clear interrupt flag (by overwriting it)
    - TWSTA: set start condition bit
    - TWEN: enable i2c-interface (pins controlled by hardware)
    */
    TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);
    loop_until_bit_is_set(TWCR, TWINT);
    return (TW_STATUS == TW_START ? 0 : 1); // 0 = ok
}


uint8_t i2c_repeatedStart(void)
{
    TWCR = (1 << TWINT) | (1 << TWSTA) | (1 << TWEN);
    loop_until_bit_is_set(TWCR, TWINT);
    return (TW_STATUS == TW_REP_START ? 0 : 1);
}


uint8_t i2c_stop(void)
{
    // TWSTO: set stop condition bit
    TWCR = (1 << TWINT) | (1 << TWSTO) | (1 << TWEN);
    // wait for stop-condition-execution (do not start before this is done!)
    loop_until_bit_is_clear(TWCR, TWSTO);
    return 0;
}


// Address a slave (connection already started)
uint8_t i2c_address(uint8_t address, uint8_t read)
{
    // Set read/write bit
    uint8_t status = i2c_write(address | (read ? 1 : 0));
    return (status == TW_MT_SLA_ACK || status == TW_MR_SLA_ACK ? 0 : 1);
}


// write byte (connection already started)
uint8_t i2c_writeByte(uint8_t data)
{
    uint8_t status = i2c_write(data);
    return (status == TW_MT_DATA_ACK || status == TW_ST_DATA_ACK ? 0 : 1);
}


// write array of bytes (connection already started)
// DO NOT use this function if the slave (eg. eeprom) needs TIME to process each written byte!
uint8_t i2c_writeData(uint8_t *data, uint8_t length)
{
    while (length--)
    {
        if (i2c_writeByte(*data))
            return 1;
        data++; // next byte
    }
    return 0;
}


// read byte (connection already started)
uint8_t i2c_readByte(uint8_t *data)
{
    return i2c_readData(data, 1);
}


// Read an array of bytes (connection already started)
uint8_t i2c_readData(uint8_t *data, uint8_t length)
{
    uint8_t status;

    // Read data
    while (length--)
    {
        // Not sending "ACK" means "this is the last byte"
        TWCR = (1 << TWINT) | (1 << TWEN) | (length > 0 ? 1 << TWEA : 0);
        loop_until_bit_is_set(TWCR, TWINT);
        status = TW_STATUS;
        if ((length && (status == TW_MR_DATA_ACK || status == TW_SR_DATA_ACK))
                    || status == TW_MR_DATA_NACK || status == TW_SR_DATA_NACK)
        {
            *data = TWDR;
            // Increase pointer
            data++;
        }
        else
            return 1;
    }
    return 0;
}


int8_t i2c_transaction(uint8_t i2cAddr, uint8_t w_length, uint8_t *w_data, uint8_t r_length, uint8_t *r_data)
{
    // write
    if (w_length > 0)
    {
        if (i2c_start() || i2c_address(i2cAddr, 0))
            goto error;
        if (i2c_writeData(w_data, w_length))
            goto error;
    }
    // read
    if (r_length > 0)
    {
        // if we wrote some data and expect some data to read, do a repeated start
        if ((w_length > 0 ? i2c_repeatedStart() : i2c_start()) || i2c_address(i2cAddr, 1))
            goto error;
        if (i2c_readData(r_data, r_length))
            goto error;
    }
    i2c_stop();
    return 0;

error:
    i2c_stop();
    return -1;
}
