#include <string.h>
#include <stdint.h>
#include "i2c.h"
#include "lm75.h"

/*
Interface LM75 like i2c-temperature sensor.

Write register: send register pointer and data
Read register: send register pointer, receive data

Device      Resolution  Accuracy    Conv.-Time  Supply      Manufacturer
LM75         9 bit      +-2.0°C      150ms      3.0-5.5V    National Semiconductor
DS75        12 bit      +-2.0°C     1200ms      2.7-5.5V    Dallas Semiconductor
DS1631      12 bit      +-0.5°C      750ms      2.7-5.5V    Dallas Semiconductor
DS7505      12 bit      +-0.5°C      200ms      1.7-3.7V    Maxim Integrated Products
TCN75        9 bit      +-0.5°C       55ms      2.7-5.5V    Microchip
TMP100      12 bit      +-2.0°C       55ms      2.7-5.5V    Texas Instruments
TMP275      12 bit      +-0.5°C      220ms      2.7-5.5V    Texas Instruments

Last change: 2015-04-19 (Frank Baettermann)
*/

// Supply pointer to 2 bytes of reserved memory
int8_t lm75_read_temperature(uint8_t i2cAddress, uint8_t *data)
{
    uint8_t reg = LM75_REG_TEMPERATURE;
    return i2c_transaction(i2cAddress, 1, &reg, 2, data);
}


// Returns temperature in deg C * 100; LM75_ERRVAL_CELSIUS on error
int16_t lm75_read_temperature_celsius(uint8_t i2caddr)
{
    uint8_t buffer[2];
    if (lm75_read_temperature(i2caddr, buffer))
        return LM75_ERRVAL_CELSIUS; // Error
    else
    {
        int16_t t = (int8_t)buffer[0];
        uint16_t t_right = (uint8_t)buffer[1];
        return t * 100 + (t_right * 100) / 256;
    }
}


// Returns temperature in deg K * 100; LM75_ERRVAL_KELVIN on error
uint16_t lm75_read_temperature_kelvin(uint8_t i2caddr)
{
    int16_t tc = lm75_read_temperature_celsius(i2caddr);

    if (tc == LM75_ERRVAL_CELSIUS)
        return LM75_ERRVAL_KELVIN; // Error
    else
        return (uint16_t)LM75_ZERODEGCINDEGK + tc;
}


int8_t lm75_read_configuration(uint8_t i2caddr, uint8_t *data)
{
    uint8_t reg = LM75_REG_CONFIGURATION;
    return i2c_transaction(i2caddr, 1, &reg, 1, data);
}


int8_t lm75_write_configuration(uint8_t i2caddr, uint8_t config)
{
    uint8_t data[2] = {LM75_REG_CONFIGURATION, config};
    return i2c_transaction(i2caddr, 2, data, 0, (uint8_t *)0);
}


#ifdef LM75_ALLREGISTERS
// Supply pointer to 2 bytes of reserved memory
int8_t lm75_read_tlow(uint8_t i2caddr, uint8_t *temp)
{
    uint8_t reg = LM75_REG_TLOW;
    return i2c_transaction(i2caddr, 1, &reg, 2, temp);
}


int8_t lm75_write_tlow(uint8_t i2caddr, uint8_t *temp)
{
    uint8_t data[3] = {LM75_REG_TLOW, temp[0], temp[1]};
    return i2c_transaction(i2caddr, 3, data, 0, (uint8_t *)0);
}


// Supply pointer to 2 bytes of reserved memory
int8_t lm75_read_thigh(uint8_t i2caddr, uint8_t *temp)
{
    uint8_t reg = LM75_REG_THIGH;
    return i2c_transaction(i2caddr, 1, &reg, 2, temp);
}


int8_t lm75_write_thigh(uint8_t i2caddr, uint8_t *temp)
{
    uint8_t data[3] = {LM75_REG_THIGH, temp[0], temp[1]};
    return i2c_transaction(i2caddr, 3, data, 0, (uint8_t *)0);
}
#endif // LM75_ALLREGISTERS
