#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <util/atomic.h>
#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "adc.h"
#include "hd44780.h"
#include "i2c.h"
#include "lm75.h"
#include "pcf8574.h"
#include "time.h"
#include "uart.h"

/*
Mobile device charge station

Target: ATMega88/168/328
Fuses:  efuse: 0x01, hfuse: 0xdc, lfuse: 0xdf

20150427 Frank Baettermann (frank@ich-war-hier.de)

- After pressing the corresponding button for a charging slot is pressed, it is
  activated for a configured amount of time. Pressing it again will extend the
  time again to the full value.
- The battery voltage limits are not altered by temperature, as the latter is
  a major factor to (usable) capacity only.
- Limping mode: Entered if we do enable new slots. If battery voltage drops further
  we drop one active slot after another trying to keep up the battery voltage.
- To calibrate current readouts connect serial port (9600 baud, 8N1) and
  turn pots to match the measured current to one measured with external multimeter.


Possible improvements (might need more flash):
- Log runtime, energy charged, etc... to EEPROM and output though UART during startup
- Better layout, use higher pincount controller to avoid porte xpanders.
*/


status_t status;

uint16_t panel_voltage = 0; // [mV]
uint16_t panel_current = 0; // [mA]
uint32_t panel_power = 0; // [mW]
uint16_t battery_voltage = 0; // [mV]
uint16_t load_current = 0; // [mA]
uint32_t load_power = 0; // [mW]
int16_t temperature; // [°C * 100]

chrg_slot_t slot[CHRG_SLOTS];

lcdbuffer_t lcd_buffer0;
lcdbuffer_t lcd_buffer1;
lcdbuffer_t* lcd_buffer_show = &lcd_buffer0;
lcdbuffer_t* lcd_buffer_draw = &lcd_buffer1;
volatile uint8_t lcd_buffer_show_write = 0; // >0 if the "show" buffer is not fully written to display, yet


// Switch "draw"/"show" pointers, draw new "show" buffer to display
void lcd_draw(void)
{
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        lcdbuffer_t* swap = lcd_buffer_show;
        lcd_buffer_show = lcd_buffer_draw;
        lcd_buffer_draw = swap;
    }
    // The new "show" buffer is written to the display somewhere else.
    // The cursor coordinates are used to remember where to continue
    // if the writing is interrupted.
    hd44780_set_cursor(0, 0);
    lcd_buffer_show->cursorx = 0;
    lcd_buffer_show->cursory = 0;
    lcd_buffer_show_write = 1;
}


// Only for testing purposes: Complete "show" buffer is written to display.
void lcd_flush(void)
{
    // Overwriting without clearing prevents flickering
    for (uint8_t y = 0; y < LCD_ROWS; y++)
    {
        hd44780_set_cursor(0, y);
        for (uint8_t x = 0; x < LCD_COLS; x++)
        {
            hd44780_putc(lcd_buffer_show->data[y][x]);
        }
    }
    lcd_buffer_show_write = 0;
}


// Clear current "draw" buffer
void lcd_clear(void)
{
    memset(lcd_buffer_draw->data, ' ', LCD_ROWS*LCD_COLS);
    lcd_buffer_draw->cursorx = 0;
    lcd_buffer_draw->cursory = 0;
}


void lcd_set_cursor(uint8_t x, uint8_t y)
{
    lcd_buffer_draw->cursorx = x;
    lcd_buffer_draw->cursory = y;
}


// Write character to lcd buffer
void lcd_putc(char c)
{
    if (lcd_buffer_draw->cursorx < LCD_COLS)
        lcd_buffer_draw->data[lcd_buffer_draw->cursory][lcd_buffer_draw->cursorx++] = c;
}


// Write zero terminated string
void lcd_puts(char *s)
{
    uint8_t x = lcd_buffer_draw->cursorx;
    uint8_t y = lcd_buffer_draw->cursory;

    while (x < LCD_COLS && *s)
    {
        lcd_buffer_draw->data[y][x++] = *s++;
    }
    lcd_buffer_draw->cursorx = x;
}


// Write zero terminated string from pgmspace
void lcd_puts_P(const char *s)
{
    uint8_t x = lcd_buffer_draw->cursorx;
    uint8_t y = lcd_buffer_draw->cursory;

    while (x < LCD_COLS)
    {
        char c = pgm_read_byte(s++);
        if (!c)
            break;
        lcd_buffer_draw->data[y][x] = c;
        x++;
    }
    lcd_buffer_draw->cursorx = x;
}


void uart_print_header(void)
{
    char buf[17];

    uart_puts_P(PSTR("Runtime [s]"));

    uart_putc(CSV_DELIMITER);
    uart_puts_P(PSTR("Panel_Voltage [mV]"));
    uart_putc(CSV_DELIMITER);
    uart_puts_P(PSTR("Panel_Current [mA]"));
    uart_putc(CSV_DELIMITER);
    uart_puts_P(PSTR("Panel_Power [mW]"));
    uart_putc(CSV_DELIMITER);
    uart_puts_P(PSTR("Battery_Voltage [mV]"));
    uart_putc(CSV_DELIMITER);
    uart_puts_P(PSTR("Load_Current [mA]"));
    uart_putc(CSV_DELIMITER);
    uart_puts_P(PSTR("Load_Power [mW]"));
    uart_putc(CSV_DELIMITER);
    uart_puts_P(PSTR("Temperature [°C*100]"));

    for (uint8_t i = 0; i < CHRG_SLOTS; i++)
    {
        uart_putc(CSV_DELIMITER);
        ultoa(i, buf, 10);
        uart_puts_P(PSTR("Slot_"));
        uart_puts(buf);
        uart_puts_P(PSTR("_Timeout [s]"));
    }

    uart_puts("\n");
}


void uart_print(void)
{
    char buf[17];

    runtime_t rt;
    time_get_runtime(&rt);
    uart_puts(ultoa(rt.s, buf, 10));
    uart_putc(CSV_DELIMITER);
    uart_puts(ultoa(panel_voltage, buf, 10));
    uart_putc(CSV_DELIMITER);
    uart_puts(ultoa(panel_current, buf, 10));
    uart_putc(CSV_DELIMITER);
    uart_puts(ultoa(panel_power, buf, 10));
    uart_putc(CSV_DELIMITER);
    uart_puts(ultoa(battery_voltage, buf, 10));
    uart_putc(CSV_DELIMITER);
    uart_puts(ultoa(load_current, buf, 10));
    uart_putc(CSV_DELIMITER);
    uart_puts(ultoa(load_power, buf, 10));
    uart_putc(CSV_DELIMITER);
    uart_puts(ltoa(temperature, buf, 10));

    for (uint8_t i = 0; i < CHRG_SLOTS; i++)
    {
        uart_putc(CSV_DELIMITER);
        uart_puts(ultoa(slot[i].timeout, buf, 10));
    }

    uart_puts("\n");
}


// Define custon chars for status display
void lcd_custom_chars_status(void)
{
    static const uint8_t chars_P[LCD_CUSTOMCHARS_STATUS][HD44780_USERCHAR_LENGTH] PROGMEM =
    {
        {0x00, 0x15, 0x0e, 0x1f, 0x0e, 0x15, 0x00, 0x00}, // LCD_CHAR_SUN
        {0x0e, 0x1b, 0x11, 0x13, 0x17, 0x1d, 0x19, 0x1f}, // LCD_CHAR_BATTERY
        {0x18, 0x18, 0x06, 0x09, 0x08, 0x08, 0x09, 0x06}  // LCD_CHAR_DEGC
    };
    // Copy each char to RAM and then send it to display
    uint8_t buf[HD44780_USERCHAR_LENGTH];
    for (uint8_t i = 0; i < LCD_CUSTOMCHARS_STATUS; i++)
    {
        memcpy_P(buf, chars_P[i], HD44780_USERCHAR_LENGTH);
        hd44780_define_char(i, buf);

        // TODO: This actual LCD seems to require another command in between
        // character definitions - or is this a problem in the hd44780/pcf8574
        // implementation?
        hd44780_clear();
    }
}


// Print status message
void lcd_status(void)
{
    char buf[11];

    lcd_clear();

    lcd_putc(LCD_CHAR_SUN);
    lcd_putc(' ');
    if (status.panel_connected)
    {
        if (panel_power < 10000)
            lcd_putc(' ');
        lcd_puts(ultoa(panel_power/1000, buf, 10));
        lcd_putc('.');
        lcd_puts(ultoa((panel_power%1000) / 100, buf, 10));
        lcd_putc('W');
    }
    else
        lcd_puts_P(PSTR("(full)"));

    // Temperature
    if (status.temperature_initialized)
    {
        int16_t tl = temperature / 100;
        int8_t tr = (temperature % 100) / 10;
        if (tr < 0)
            tr = -tr;
        ltoa(tl, buf, 10);
        lcd_set_cursor(13 - strlen(buf), 0);
        lcd_puts(buf);
        lcd_putc('.');
        lcd_putc('0' + tr);
        lcd_putc(LCD_CHAR_DEGC);
    }

    lcd_set_cursor(0, 1);
    lcd_putc(LCD_CHAR_BATTERY);
    lcd_putc(' ');
    if (status.load_connected)
    {
        if (load_power < 10000)
            lcd_putc(' ');
        lcd_puts(ultoa(load_power/1000, buf, 10));
        lcd_putc('.');
        lcd_puts(ultoa((load_power%1000) / 100, buf, 10));
        lcd_putc('W');

        if (status.limping_mode)
        {
            lcd_set_cursor(9, 1);
            lcd_puts_P(PSTR("limping"));
        }
    }
    else
        lcd_puts_P(PSTR("Battery low"));

    lcd_draw();
}


void lcd_msg_load_dropped(void)
{
    lcd_clear();
    lcd_puts_P(PSTR(" Load droppped"));
    lcd_set_cursor(3, 1);
    lcd_puts_P(PSTR("(battery low)"));
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_load_connected(void)
{
    lcd_clear();
    lcd_puts_P(PSTR(" Load connected"));
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_limping_started(void)
{
    lcd_clear();
    lcd_puts_P(PSTR("Started limping."));
    lcd_set_cursor(3, 1);
    lcd_puts_P(PSTR("(battery low)"));
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_limping_stopped(void)
{
    lcd_clear();
    lcd_puts_P(PSTR("Stopped limping."));
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_limping_slotdisable(uint8_t i)
{
    char buf[5];
    lcd_clear();
    lcd_puts_P(PSTR("Limping disabled"));
    lcd_set_cursor(0, 1);
    lcd_puts_P(PSTR("slot "));
    lcd_puts(ultoa(i + 1, buf, 10));
    lcd_puts_P(PSTR(". Sorry."));
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_limping_na(void)
{
    lcd_clear();
    lcd_puts_P(PSTR("Not available."));
    lcd_set_cursor(7, 1);
    lcd_puts_P(PSTR("(limping)"));
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_batterylow_na(void)
{
    lcd_clear();
    lcd_puts_P(PSTR("Not available."));
    lcd_set_cursor(3, 1);
    lcd_puts_P(PSTR("(battery low)"));
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_slot_enabled(uint8_t i)
{
    char buf[5];
    lcd_clear();
    lcd_puts_P(PSTR(" Slot "));
    lcd_puts(ultoa(i + 1, buf, 10));
    lcd_puts_P(PSTR(" active"));
    lcd_set_cursor(0, 1);
    lcd_puts_P(PSTR(" for "));
#if (CHRG_TIMEOUT >= 3600)
    lcd_puts(ultoa(CHRG_TIMEOUT / 3600, buf, 10));
    lcd_puts_P(PSTR(" hour(s)"));
#elif (CHRG_TIMEOUT >= 60)
    lcd_puts(ultoa(CHRG_TIMEOUT / 60, buf, 10));
    lcd_puts_P(PSTR("minute(s)"));
#else
    lcd_puts(ultoa(CHRG_TIMEOUT, buf, 10));
    lcd_puts_P(PSTR(" seconds"));
#endif
    lcd_draw();
    lcd_msg_timers();
}


void lcd_msg_slot_timedout(uint8_t i)
{
    char buf[5];
    lcd_clear();
    lcd_puts_P(PSTR("Slot "));
    lcd_puts(ultoa(i + 1, buf, 10));
    lcd_puts_P(PSTR(" has"));
    lcd_set_cursor(0, 1);
    lcd_puts_P(PSTR("timed out."));
    lcd_draw();
    lcd_msg_timers();
}


// Prevent "status" from overwriting current message
void lcd_msg_timers(void)
{
    lcd_backlight();
    time_delay_lcdmsg_s = LCD_MESSAGE_TIMEOUT;
}


void lcd_backlight(void)
{
    hd44780_pcf8574_backlight_on();
    time_delay_lcdbl_s = LCD_BACKLIGHT_TIMEOUT;
}


void slot_enable(uint8_t i)
{
    switch (i)
    {
        case 0: IO_CHAN00_PORT |= (1 << IO_CHAN00_BIT); break;
        case 1: IO_CHAN01_PORT |= (1 << IO_CHAN01_BIT); break;
        case 2: IO_CHAN02_PORT |= (1 << IO_CHAN02_BIT); break;
        case 3: IO_CHAN03_PORT |= (1 << IO_CHAN03_BIT); break;
        case 4: IO_CHAN04_PORT |= (1 << IO_CHAN04_BIT); break;
        case 5: IO_CHAN05_PORT |= (1 << IO_CHAN05_BIT); break;
        case 6: IO_CHAN06_PORT |= (1 << IO_CHAN06_BIT); break;
        case 7: IO_CHAN07_PORT |= (1 << IO_CHAN07_BIT); break;
        default:
            break;
    }

    if (i < CHRG_SLOTS)
    {
        slot[i].timeout = CHRG_TIMEOUT;
        slot[i].active = 1;
    }
}


void slot_disable(uint8_t i)
{
    switch (i)
    {
        case 0: IO_CHAN00_PORT &= ~(1 << IO_CHAN00_BIT); break;
        case 1: IO_CHAN01_PORT &= ~(1 << IO_CHAN01_BIT); break;
        case 2: IO_CHAN02_PORT &= ~(1 << IO_CHAN02_BIT); break;
        case 3: IO_CHAN03_PORT &= ~(1 << IO_CHAN03_BIT); break;
        case 4: IO_CHAN04_PORT &= ~(1 << IO_CHAN04_BIT); break;
        case 5: IO_CHAN05_PORT &= ~(1 << IO_CHAN05_BIT); break;
        case 6: IO_CHAN06_PORT &= ~(1 << IO_CHAN06_BIT); break;
        case 7: IO_CHAN07_PORT &= ~(1 << IO_CHAN07_BIT); break;
        default:
            break;
    }

    if (i < CHRG_SLOTS)
    {
        slot[i].timeout = 0;
        slot[i].active = 0;
    }
}


// Returns the number of currently active slots
uint8_t slots_active(void)
{
    uint8_t count = 0;
    for (uint8_t i = 0; i < CHRG_SLOTS; i++)
    {
        if (slot[i].active)
            count++;
    }
    return count;
}


// Returns index of the slot with lowest timeout value (was longer active than the other)
// If there is no active slot it returns UINT8_MAX
uint8_t slot_longestactive(void)
{
    uint8_t ret = UINT8_MAX;
    uint16_t lowest_timeout = UINT16_MAX;
    for (uint8_t i = 0; i < CHRG_SLOTS; i++)
    {
        if (slot[i].active && slot[i].timeout < lowest_timeout)
        {
            ret = i;
            lowest_timeout = slot[i].timeout;
        }
    }


    return ret;
}


void main_init(void)
{
    // Outputs
    IO_CHARGEREN_DDR |= (1 << IO_CHARGEREN_BIT);
    IO_CHARGEREN_UNSET;
    IO_ALLOUTEN_DDR |= (1 << IO_ALLOUTEN_BIT);
    IO_ALLOUTEN_UNSET;
    IO_ALLOUTINRUSH_DDR |= (1 << IO_ALLOUTINRUSH_BIT);
    IO_LEDFULL_DDR |= (1 << IO_LEDFULL_BIT);
    // Charger channels
    IO_CHAN00_DDR |= (1 << IO_CHAN00_BIT);
    IO_CHAN01_DDR |= (1 << IO_CHAN01_BIT);
    IO_CHAN02_DDR |= (1 << IO_CHAN02_BIT);
    IO_CHAN03_DDR |= (1 << IO_CHAN03_BIT);
    IO_CHAN04_DDR |= (1 << IO_CHAN04_BIT);
    IO_CHAN05_DDR |= (1 << IO_CHAN05_BIT);
    IO_CHAN06_DDR |= (1 << IO_CHAN06_BIT);
    IO_CHAN07_DDR |= (1 << IO_CHAN07_BIT);

    // Status
    status.load_connected = 0;
    status.panel_connected = 0;
    status.limping_mode = 0;
    status.led_status = 0;
    status.temperature_initialized = 0;

    // Charger channels
    for (uint8_t i = 0; i < CHRG_SLOTS; i++)
    {
        slot[i].enabled = 1;
        slot[i].active = 0;
        slot[i].timeout = 0;
    }

    // Internal peripherals
    time_init();
    uart_init();
    adc_init();
    adc_enable();
    adc_start_conversion();
}


void main_buttons(void)
{
    static uint8_t buttons_last = 0xff;

    uint8_t b;
    if (!pcf8574_read(BUTTON_PCF8574_I2CADDR, &b))
    {
        // Detect rising edges (end of button press)
        uint8_t mask = 0x01;
        for (uint8_t i = 0; i < CHRG_SLOTS; i++)
        {
            if (b & mask && !(buttons_last & mask) && slot[i].enabled)
            {
                if (!time_delay_lcdbl_s)
                    lcd_backlight(); // First button press enables backlight
                else if (status.limping_mode)
                    lcd_msg_limping_na();  // N/A: Do not enable/extend (limping)
                else if (!status.load_connected)
                    lcd_msg_batterylow_na(); // N/A: Load already dropped
                else
                {
                    // Tolerate inrush current while load switching
                    time_delay_loadinrush_ms = AUIR3325_INRUSH_DELAY;
                    IO_ALLOUTINRUSH_SET;

                    slot_enable(i);
                    lcd_msg_slot_enabled(i);
                }
                break; // We evaluate only one button at at time...
            }
            mask <<= 1;
        }
        buttons_last = b;
    }
}


void main_measurements(void)
{
    // Measurements using internal ADC
    if (adc_voltage_new && adc_current_new)
    {
        // Copy values
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
        {
            panel_voltage = adc_voltage;
            adc_voltage_new = 0;
            panel_current = adc_current;
            adc_current_new = 0;
        }
        // Scale panel voltage (100k/10k divider = *11)
        panel_voltage = ((uint32_t)ADC_REFVOLTAGE_AREF * panel_voltage * 11UL) / ADC_SS_MAX;
        // Scale current (15k/15k divider = *2; adjust trimpot to compensate errors)
        // U_Rifb = adc_value * U_Ref * 2 (divider) / adc_max
        // I = U_Rifb * 2800 / Rifb
        panel_current = ((uint64_t)panel_current * ADC_REFVOLTAGE_AREF * 2 * AUIR3315_CURRENTRATIO) / ((uint32_t)ADC_SS_MAX * AUIR3315_PANEL_RIFB);
        panel_power = ((uint32_t)panel_voltage * panel_current) / 1000;
    }
    if (adc_battery_voltage_new && adc_load_current_new)
    {
        // Copy values
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
        {
            battery_voltage = adc_battery_voltage;
            adc_battery_voltage_new = 0;
            load_current = adc_load_current;
            adc_load_current_new = 0;
        }
        // Scale battery voltage (100k/10k divider = *11)
        battery_voltage = ((uint32_t)ADC_REFVOLTAGE_AREF * battery_voltage * 11UL) / ADC_SS_MAX;
        // Scale current (15k/15k divider = *2; adjust trimpot to compensate errors)
        load_current = ((uint64_t)load_current * ADC_REFVOLTAGE_AREF * 2 * AUIR3315_CURRENTRATIO) / ((uint32_t)ADC_SS_MAX * AUIR3315_LOAD_RIFB);
        load_power = ((uint32_t)battery_voltage * load_current) / 1000;
    }
}


void main_measurements_temperature(void)
{
    // This uses the common 250ms interval

    if (!status.temperature_initialized)
        status.temperature_initialized = lm75_write_configuration(LM75_I2CADDR, TMP275_CONF_RES12) ? 0 : 1;
    // Now try to read temperature
    if (status.temperature_initialized)
    {
        temperature = lm75_read_temperature_celsius(LM75_I2CADDR);
        if (temperature == LM75_ERRVAL_CELSIUS)
            status.temperature_initialized = 0;
    }
}


// Control connection of battery charger (PV panel) and load (charing slots)
void main_powerctrl(void)
{
    uint8_t slot_disabled = 0;

    // LED "full": indicate full state if voltage is greater than reacharge threshold
    if (battery_voltage > BATT_VOLTAGE_RECHARGE)
        IO_LEDFULL_SET;
    else
        IO_LEDFULL_UNSET;

    // Connect/disconnect charger (with hysteresis)
    if (status.panel_connected && battery_voltage >= BATT_VOLTAGE_STOPCHARGE)
    {
        status.panel_connected = 0;
        IO_CHARGEREN_UNSET;
    }
    else if (!status.panel_connected && battery_voltage <= BATT_VOLTAGE_RECHARGE)
    {
        status.panel_connected = 1;
        IO_CHARGEREN_SET;
    }

    // If charging slot times out: deactivate
    for (uint8_t i = 0; i < CHRG_SLOTS; i++)
    {
        if (slot[i].active)
        {
            if (slot[i].timeout > MAIN_POWERCTRL_INTERVAL)
                slot[i].timeout -= MAIN_POWERCTRL_INTERVAL;
            else
            {
                slot_disabled++;
                slot_disable(i);
                lcd_msg_slot_timedout(i);
            }
        }
    }

    // Control limping mode (does not accept new slots)
    if (status.load_connected)
    {
        if (status.limping_mode)
        {
            if (battery_voltage >= BATT_VOLTAGE_LIMPINGMODE_STOP)
            {
                // Possibility to leave limping mode without dropping the load in the mean time
                status.limping_mode = 0;
                lcd_msg_limping_stopped();
            }
            else if (battery_voltage <= BATT_VOLTAGE_LIMPINGMODE_DROP1 && slots_active() > 1)
            {
                // If battery voltage keeps falling and there are more than one slots
                // active drop the one which has highest charge time giving the other
                // slots more possibility to charge (hopefully)
                // The last active slot is not deactivated here, because that would equal
                // the full drop of all load.
                uint8_t i = slot_longestactive();
                if (i < CHRG_SLOTS)
                {
                    lcd_msg_limping_slotdisable(i);
                    slot_disabled++;
                    slot_disable(i);
                }
            }
        }
        else if (!status.limping_mode && battery_voltage < BATT_VOLTAGE_LIMPINGMODE_ENTER)
        {
            status.limping_mode = 1;
            lcd_msg_limping_started();
        }
    }

    // Connect/disconnect load completely
    if (status.load_connected && battery_voltage < BATT_VOLTAGE_LOAD_DROP)
    {
        // Only drop load completely if no slots we're disabled (during this function call)
        if (!slot_disabled)
        {
            // Disconnect all slots
            for (uint8_t i = 0; i < CHRG_SLOTS; i++)
                slot_disable(i);
            status.load_connected = 0;
            status.limping_mode = 0;
            IO_ALLOUTEN_UNSET;
            lcd_msg_load_dropped();
        }
    }
    else
    {
        if (!status.load_connected && battery_voltage >= BATT_VOLTAGE_LOAD_CONNECT)
        {
            // Reconnect load
            status.load_connected = 1;
            status.limping_mode = 0;
            IO_ALLOUTINRUSH_SET;
            time_delay_loadinrush_ms = AUIR3325_INRUSH_DELAY;
            IO_ALLOUTEN_SET;
            lcd_msg_load_connected();
        }
    }
}


// Write LCD buffer to display
void main_lcd(void)
{
    // Write one character per call - reduces jitter for more important stuff to do.
    if (lcd_buffer_show_write)
    {
        hd44780_putc(lcd_buffer_show->data[lcd_buffer_show->cursory][lcd_buffer_show->cursorx]);
        // Next character
        lcd_buffer_show->cursorx++;
        if (lcd_buffer_show->cursorx >= LCD_COLS)
        {
            // Next line
            lcd_buffer_show->cursory++;
            if (lcd_buffer_show->cursory >= LCD_ROWS)
                lcd_buffer_show_write = 0; // Done
            else
            {
                lcd_buffer_show->cursorx = 0;
                hd44780_set_cursor(0, lcd_buffer_show->cursory);
            }
        }
    }
}


int main(void)
{
    // Deactivate watchdog (otherwise it will remain enabled after a system reset)
    wdt_disable();
    main_init();
    i2c_init();

    // Set all button pins high (input)
    pcf8574_write(BUTTON_PCF8574_I2CADDR, 0xff);

    // LCD
    hd44780_init();
    hd44780_pcf8574_backlight_on();
    time_delay_lcdbl_s = LCD_BACKLIGHT_TIMEOUT;
    lcd_custom_chars_status();

    // Startup message
    lcd_clear();
    lcd_puts_P(PSTR(STARTUP_TEXT1));
    lcd_set_cursor(0, 1);
    lcd_puts_P(PSTR(STARTUP_TEXT2));
    lcd_draw();
    lcd_flush();
    _delay_ms(1000);

    wdt_enable(WDTO_2S);
    sei();
    uart_print_header();

    for (;;)
    {
        // Button presses to enable display/charge port
        if (!time_delay_button_ms)
        {
            time_delay_button_ms = BUTTON_INTERVAL;
            main_buttons();
        }

        if (!time_delay_uart_s)
        {
            time_delay_uart_s = 1;
            uart_print();
        }

        main_measurements();

        if (!time_delay_250ms)
        {
            time_delay_250ms = 250;

            if (!time_delay_powerctrl_s)
            {
                time_delay_powerctrl_s = MAIN_POWERCTRL_INTERVAL;
                main_powerctrl();
            }

            // Temperature
            main_measurements_temperature();

            // Update status screen (if there is no message)
            if (!time_delay_lcdmsg_s)
                lcd_status();

            if (!time_delay_lcdbl_s)
                hd44780_pcf8574_backlight_off();

//IO_ALLOUTINRUSH_UNSET;
//IO_ALLOUTINRUSH_SET;

//            // Blink status LED: show we're alive
//            if (status.led_status)
//            {
//                status.led_status = 0;
//                // No LED present :-(
//            }
//            else
//            {
//                status.led_status = 1;
//                // No LED present :-(
//            }
        }

        main_lcd();

        // Deactivate inrush current tolerance for load switch
        if (!time_delay_loadinrush_ms)
            IO_ALLOUTINRUSH_UNSET;

        // Reset watch dog
        wdt_reset();
    }
    return 0;
}
