#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include "fifo.h"
#include "main.h"
#include "uart.h"


// UART buffers
uint8_t com_uartRxFifoBuffer[UART_BUFFER_RX];
uint8_t com_uartTxFifoBuffer[UART_BUFFER_TX];
fifo_t fifoRx;
fifo_t fifoTx;


void uart_init(void)
{
    // Initialize UART buffers and UART
    fifo_init(&fifoRx, com_uartRxFifoBuffer, UART_BUFFER_RX);
    fifo_init(&fifoTx, com_uartTxFifoBuffer, UART_BUFFER_TX);

    // deactivate interrupts temporarily
    uint8_t sreg = SREG;
    cli();

    // UCSRnA – USART Control and Status Register n A
    // nothing to do

    // UCSRnB – USART Control and Status Register n B
    UCSR0B = 1 << RXCIE0 // RX Complete Interrupt Enable
           | 1 << UDRIE0 // USART Data Register Empty Interrupt Enable
           | 1 << RXEN0  // Enable Receiver
           | 1 << TXEN0; // Enable Transmitter

    // UCSRnC – USART Control and Status Register n C
    UCSR0C = 1 << UCSZ01 | 1 << UCSZ00; // asynchronous 8N1

    // UBRRnL and UBRRnH - USART Baud Rate Registers
    UBRR0 = UART_UBRR_VAL;

    // Flush receive-buffer
    do
    {
        UDR0; // actual value unused
    }
    while (UCSR0A & (1 << RXC0));

    // Restore interrupt settings (enable interrupts somewhere else)
    SREG = sreg;

    // Port directions
    DDRD &= ~(1 << PD0); // UART0 RX
    DDRD |= (1 << PD1); // UART0 TX
}


// Save received byte in receive fifo
ISR(USART_RX_vect)
{
    fifo_put(&fifoRx, UDR0);
}


// Get byte from send buffer. When byte has been sent a new interrupt will be
// triggered and deactivated if send fifo is empty. The interrupt will be
// reactivated when we write to send buffer.
ISR(USART_UDRE_vect)
{
    if (fifoTx.count)
        UDR0 = fifo_get(&fifoTx);
    else
        UCSR0B &= ~(1 << UDRIE0);
}


// Read byte from receive buffer.
// Returns -1 if buffer is empty.
int16_t uart_getc(void)
{
    return fifo_get(&fifoRx);
}


// Wait until we receive one byte and return it.
uint8_t uart_getc_wait(void)
{
    return fifo_get_wait(&fifoRx);
}


// Write single character to send buffer.
void uart_putc(const char c)
{
    fifo_put_wait(&fifoTx, c);
    UCSR0B |= (1 << UDRIE0); // re-enable data register empty interrupt
}


// Write null terminated string to send buffer.
void uart_puts(char *s)
{
    for (;;)
    {
        if (!*s) // null termination
            break;
        fifo_put_wait(&fifoTx, *s++);
        UCSR0B |= (1 << UDRIE0);
    }
}


// Transmit byte array
void uart_put(uint8_t *a, uint8_t len)
{
    while (len--)
    {
        uart_putc(*a);
        a++;
    }
}


// Write null terminated string from flash to send buffer.
void uart_puts_P(const char *s)
{
    for (;;)
    {
        char c = pgm_read_byte(s++);
        if (!c)
            break;
        fifo_put_wait(&fifoTx, c);
        UCSR0B |= (1 << UDRIE0);
    }
}
